#define _XOPEN_SOURCE /* glibc2 needs this */

#include "config.h"

#include "epg.hpp"

#include "input.hpp"
#include "common.hpp"
#include "graphics.hpp"
#include "updater.hpp"
#include "common-feature.hpp"
#include "startmenu.hpp"
#include "busy_indicator.hpp"
#include "shutdown.hpp"
#include "extra_menu.hpp"
#include "touch.hpp"

#include "search.hpp"

#include "plugins.hpp"

#include <time.h>

#include "gettext.hpp"

#include "boost.hpp"

#include <iostream>
#include <list>
#include <sstream>
#include <algorithm>

// time of last change
#include <sys/stat.h>

#include "tinyxml.h"

#include "svdrp.hpp"

#ifdef use_nls
#include <langinfo.h>
#include <locale.h>
#endif

using std::string;
using std::vector;
using std::list;

pthread_mutex_t Epg::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

time_t Event::str_to_time(const char *s) {
  struct tm tm;
  char * datetimetz;

#ifdef __UCLIBC__
  char Y[5], m[3], d[3], H[3], M[3];
  char t[32];

  s += 0; strncpy(Y, s, 4); Y[4] = 0;
  s += 4; strncpy(m, s, 2); m[2] = 0;
  s += 2; strncpy(d, s, 2); d[2] = 0;
  s += 2; strncpy(H, s, 2); H[2] = 0;
  s += 2; strncpy(M, s, 2); M[2] = 0;
  sprintf(t, "%s-%s-%s %s:%s:00 +0000", Y, m, d, H, M);
#endif

  int hours=0;
  int minutes=0;
  int factor=1;

  //fix unset value
  tm.tm_isdst=0;
  //get timezone
  tzset();

#ifdef __UCLIBC__
  datetimetz = strptime(t, "%Y-%m-%d %H:%M:%S", &tm);
#else
  datetimetz = strptime(s, "%Y%m%d%H%M%S", &tm);
#endif

#if 0

  //check if data reliable
  if (datetimetz==NULL || strlen(datetimetz)!=6)
    return-1;
  //move up space
  datetimetz++;
  //get factor
  if (datetimetz[0]=='-') {
    factor=-1;
  }
  //move up sign
  datetimetz++;
  //copy XXXX to extract hours
  char tmptz[5];
  strcpy(tmptz,datetimetz);
  //we only need hours
  tmptz[2]=0;
  //get hours
  hours = atoi(tmptz);
  //get minutes
  minutes= atoi(&datetimetz[2]);
  //correction using local timezone and xml timezone to obtain the
  //right date time
	tm.tm_sec+=timezone + factor*((minutes * SECONDS_PER_MINUTE)+(hours * 3600));
#endif

  return mktime(&tm);
}

bool Event::operator<(const Event& rhs) const
{
  return (this->start < rhs.start);
}

Event::Event() {
	start = 0;
	stop = 0;
	duration = 0;
}

Event::~Event() {
	
}

EpgEvent::EpgEvent() : Event(){
}

EpgEvent::~EpgEvent() {
	
}

RObj *XmlEpgEvent::Print(Render *render, Themes *themes, int x, int y, int width, int height, bool selected, std::string fontname) {
	RObj *r;
	string name = title;
	string_format::format_to_size(name, fontname, width - 10, false);

	if (selected) {
		r = new RObj(x, y, width - 5, height,
				themes->epg_marked1,
				themes->epg_marked2,
				themes->epg_marked3,
				themes->epg_marked4, 3);
	
		render->current.add(r);
		render->current.add(new TObj(name, fontname, x + 2, y,
				themes->epg_marked_font1,
				themes->epg_marked_font2,
				themes->epg_marked_font3, 4));
	} else {
		r = new RObj(x, y, width - 5, height,
				themes->epg_boxes1,
				themes->epg_boxes2,
				themes->epg_boxes3,
				themes->epg_boxes4, 3);
	
		render->current.add(r);
		render->current.add(new TObj(name, fontname, x + 2, y,
				themes->epg_boxes_font1,
				themes->epg_boxes_font2,
				themes->epg_boxes_font3, 4));
	}
	return r;
}

void XmlEpgEvent::PrintDescription(Render *render, Themes *themes, int width, int height) {
	char time_str[20];
	std::ostringstream time;
	
	InputMaster *input_master = S_InputMaster::get_instance();
	Epg *epg = get_class<Epg>(dgettext("mms-epg", "Electronic Program Guide"));
	
	int i = 0;
	int round = 1;
	string::iterator it = desc.begin();
	std::ostringstream next_word;
	
	int y = BORDER_TOP;
	
	while (i < desc.size()) {
	
		render->prepare_new_image();
	
		render->current.add(new PObj(themes->epg_background, 0, 0, 0, SCALE_FULL));
	
		render->current.add(new PFObj(themes->startmenu_epg, BORDER_LEFT, y, epg->header_box_size, epg->header_box_size, 1, true));
	
		render->current.add(new TObj("EPG", epg->header_font, BORDER_LEFT + epg->header_box_size + SPACING, y + (epg->header_box_size - epg->header_size.second) / 2,
				themes->epg_header_font1, themes->epg_header_font2,
				themes->epg_header_font3, 1));
	
		y += epg->header_size.second + SPACING;
		
		render->current.add(new RObj(BORDER_LEFT, y, width - BORDER_LEFT - BORDER_RIGHT, height - y - BORDER_BOTTOM,
				themes->general_rectangle_color1, themes->general_rectangle_color2,
				themes->general_rectangle_color3, themes->general_rectangle_color4, 2));
				
		y += SPACING;
	
		int x_off = 0;
		
		if (epg->display_channel_logo && !epg->sel_channel->logo.empty()) {
			render->create_scaled_image_wrapper_upscaled(epg->sel_channel->logo, epg->title_size.second, epg->title_size.second);
			render->current.add(new PFObj(epg->sel_channel->logo, BORDER_LEFT + SPACING, y,  epg->title_size.second, epg->title_size.second, false, 2));
			x_off = epg->title_size.second + SPACING;
		}
		
		if (epg->display_channel_name) {
			string_format::format_to_size(epg->sel_channel->display_name, epg->title_font, width - BORDER_LEFT - BORDER_RIGHT - SPACING + x_off, true);
					
			render->current.add(new TObj(epg->sel_channel->display_name, epg->title_font, BORDER_LEFT + SPACING + x_off, y,
						themes->epg_font1, themes->epg_font2, themes->epg_font3, 2));
					
		}
		y += epg->title_size.second;

		strftime(time_str, 20, "%H:%M", localtime(&start));
		time << time_str << " - ";
		strftime(time_str, 20, "%H:%M", localtime(&stop));
		time << time_str;
	
		render->current.add(new TObj(time.str(), epg->subtitle_font, BORDER_LEFT + SPACING, y,
				themes->epg_font1, themes->epg_font2, themes->epg_font3, 2));
		
		x_off = string_format::calculate_string_width(time.str(), epg->subtitle_font);

		time.str("");
	
		string_format::format_to_size(title, epg->subtitle_font, width - BORDER_LEFT - BORDER_RIGHT - SPACING - x_off, true);
			
		render->current.add(new TObj(title, epg->subtitle_font, BORDER_LEFT + SPACING + x_off + SPACING, y,
				themes->epg_font1, themes->epg_font2, themes->epg_font3, 2));
	
		y += epg->subtitle_size.second;
		
		int start_y = y;
		
		int lines_of_text = (height - y - BORDER_BOTTOM) / epg->normal_size.second;
		int box_width = width - BORDER_LEFT - BORDER_RIGHT - (2 * SPACING);
		int x = BORDER_LEFT + SPACING;
		
		while (i != desc.size()) {
			string descr = string_format::get_line(desc, i, box_width, epg->normal_font);
	
			render->current.add(new TObj(descr, epg->normal_font, x, y,
					themes->epg_font1, themes->epg_font2,
					themes->epg_font3, 2));
			y += epg->normal_size.second;
	
			if (y >= start_y + lines_of_text * epg->normal_size.second)
				break;
		}
	
		render->draw_and_release("EPG description");
	
		// wait for keypress
		input_master->get_input();
	
		++round;
		y = BORDER_TOP;
	}
}

XmlEpgEvent::XmlEpgEvent(TiXmlElement *element) : EpgEvent() {
  TiXmlNode *node;
  TiXmlText *text;

  if (element->Attribute("start")) {
    start = str_to_time(element->Attribute("start"));
  }

  if (element->Attribute("stop")) {
    stop = str_to_time(element->Attribute("stop"));
  }

	duration = 0;
	if (stop > start)
		duration = stop - start;
	
  if (element->FirstChild("title")) {
    node = element->FirstChild("title");
    if (node) {
      node = node->FirstChild();
      if (node) {
	text = node->ToText();
	if (text)
	  title = text->Value();
      }
    }
  }

  if (element->FirstChild("desc")) {
    node = element->FirstChild("desc");
    if (node) {
      node = node->FirstChild();
      if (node) {
	text = node->ToText();
	if (text)
	  desc = text->Value();
      }
    }
  }
}

XmlEpgEvent::XmlEpgEvent(time_t _start, long _duration, string _title) : EpgEvent() {
	start = _start;
	stop = start + _duration;
    duration = _duration;
	title = _title;
}

XmlEpgEvent::~XmlEpgEvent() {
	
}

Channel::Channel() {
	
}

Channel::~Channel() {
	std::vector<EpgEvent *>::iterator ev;
	
	// Delete all Epg-Events;
	for (ev = program.begin(); ev != program.end(); ev++) {
		delete (*ev);
	}
	program.clear();
}

XMLChannel::XMLChannel(int _number, TiXmlElement *element) : Channel()
{
  TiXmlNode *node = 0;
  TiXmlText *text = 0;
	epgsource = XML;
	number = _number;
	logo = "";

  if (element->Attribute("id")) {
    id = element->Attribute("id");
  }

  if (element->FirstChild("display-name")) {
    node = element->FirstChild("display-name");
    if (node != 0) {
      node = node->FirstChild();
      if (node != 0) {
	text = node->ToText();
	display_name = text->Value();
      }
    }
  }
}

void Epg::create_db()
{
  vector<string> fields;
  fields.push_back("version INTEGER NOT NULL");
  fields.push_back("date TEXT");
  db.createTable("Schema", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("tvid TEXT");
  fields.push_back("name TEXT");
	// VDR
	fields.push_back("number LONG");
	fields.push_back("shortname TEXT");
	fields.push_back("frequency INTEGER");
	fields.push_back("param TEXT");
	fields.push_back("source TEXT");
	fields.push_back("rate INTEGER");
	fields.push_back("vpid TEXT");
	fields.push_back("apid TEXT");
	fields.push_back("tpid INTEGER");
	fields.push_back("caid TEXT");
	fields.push_back("sid INTEGER");
	fields.push_back("nid INTEGER");
	fields.push_back("tnid INTEGER");
	fields.push_back("tid INTEGER");
	fields.push_back("rid INTEGER");
  db.createTable("Channel", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("cid TEXT"); // channel id
  fields.push_back("title TEXT");
  fields.push_back("ltitle TEXT");
  fields.push_back("description TEXT");
  fields.push_back("ldescription TEXT");
	fields.push_back("start_time LONG");
	fields.push_back("end_time LONG");
	// VDR
	fields.push_back("vps LONG");
	fields.push_back("duration LONG");
  db.createTable("Program", fields);

	// VDR-Timer - NOT YET IMPLEMENTED
	fields.clear();
	fields.push_back("id INTEGER PRIMARY KEY");
	fields.push_back("number INTEGER"); 	// Timer
	fields.push_back("flag INTEGER"); 		// Flags
	fields.push_back("channel INTEGER"); 	// channel number
	fields.push_back("days TEXT"); 			// Day of Timer Or Repetitive-Pattern
	fields.push_back("repetitive INTEGER");
	fields.push_back("start_time LONG");
	fields.push_back("end_time LONG");
	fields.push_back("duraion LONG");
	fields.push_back("priority INTEGER");
	fields.push_back("lifetime INTEGER");
	fields.push_back("filename TEXT");
	fields.push_back("info TEXT");
	db.createTable("Timer", fields);

	
  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("pid INTEGER NOT NULL");
  fields.push_back("year INTEGER NOT NULL");
  db.createTable("MovieProgram", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("pid INTEGER NOT NULL");
  fields.push_back("nr INTEGER NOT NULL");
  db.createTable("EpisodeProgram", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  fields.push_back("role TEXT");
  db.createTable("Actors", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("aid INTEGER NOT NULL");
  fields.push_back("pid INTEGER NOT NULL");
  db.createTable("AProgram", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  db.createTable("Directors", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("did INTEGER NOT NULL");
  fields.push_back("pid INTEGER NOT NULL");
  db.createTable("DProgram", fields);

  create_indexes();
}

void Epg::create_indexes()
{
  db.execute("CREATE INDEX idx_program_title ON Program (ltitle)");
  db.execute("CREATE INDEX idx_program_desc ON Program (ldescription)");
  db.execute("CREATE INDEX idx_channel_tvid ON Channel (tvid)");

	db.execute("CREATE INDEX idx_channel_number ON Channel (number)");

}

void Epg::clear_db()
{
	SQLQuery *dq = db.query("sqlite_master",
			"SELECT name FROM %t WHERE type='table'");
  if (dq) {
    for (int i = 0; i < dq->numberOfTuples(); ++i) {
      db.execute(("DROP TABLE " + (*dq)[i]["name"]).c_str());
    }
  }

  delete dq;

  dq = db.query("sqlite_master", "SELECT name FROM %t WHERE type='index'");
  if (dq) {
    for (int i = 0; i < dq->numberOfTuples(); ++i) {
      db.execute(("DROP INDEX " + (*dq)[i]["name"]).c_str());
    }
  }

  delete dq;
}

void Epg::insert_schema()
{
  struct stat64 s;
  stat64(epg_conf->p_epg_data().c_str(), &s);

	char *tmp = sqlite3_mprintf("INSERT INTO Schema VALUES('4', '%q')",
			      conv::itos(s.st_mtime).c_str());
  db.execute(tmp);
  sqlite3_free(tmp);
}

bool Epg::check_db()
{
  if (!file_exists(epg_conf->p_epg_data()))
    return true;

  bool db_ok = true;

  if (db.hasTable("Schema")) {
    // check for correct version
    SQLQuery *q = db.query("Schema", "SELECT * FROM %t");
    if (q && q->numberOfTuples() > 0) {
      SQLRow &row = (*q)[0];
      struct stat64 s;
      stat64(epg_conf->p_epg_data().c_str(), &s);
      if (row["date"] != conv::itos(s.st_mtime)) {
	clear_db();
	create_db();
	insert_schema();
	db_ok = false;
			} else if (row["version"] != "4") {

	Print print(gettext("Upgrading EPG database schema"), Print::INFO);

	clear_db();
	create_db();
	insert_schema();
      }
    } else {
      clear_db();
      create_db();
      insert_schema();
      db_ok = false;
    }
    delete q;
  } else {
    // create the whole db
    clear_db();
    create_db();
    insert_schema();
    db_ok = false;
  }

  return db_ok;
}

void Epg::update_time() {
	struct tm *tm;

	cur_time = time(0);
	tm = localtime(&cur_time);
	if (tm->tm_min < timeslot_interval) {
		tm->tm_min = 0;
	} else {
		tm->tm_min = timeslot_interval;
  }
	tm->tm_sec = 0;
	cur_time = mktime(tm);
	sel_time = cur_time;
  }

void Epg::read_db() {
	if (use_svdrp) {
		if (!GetSvdrpData())
			return;
	} else {
		if (!file_exists(epg_conf->p_epg_data()))
			return;

		if (!read_xml_file(&channels))
			return;
  }

	cur_channel = channels.begin();
	sel_channel = cur_channel;
	sel_channel_num = 0;

	struct tm *tm;

	cur_time = time(0);
	sel_time = cur_time;

	for (program_iter = sel_channel->program.begin(), program_iter_end
			= sel_channel->program.end(); program_iter != program_iter_end; ++program_iter)
		if ((*program_iter)->stop >= sel_time)
	  break;

	if (program_iter != program_iter_end) {
		cur_time = sel_time = (*program_iter)->start;
	}
}

void Epg::parse_channels(std::string _channels, std::vector<int> *_channelnumbers) {
	std::vector<int>::iterator it;

	int fc, lc, posc, posr;
	std::string range;

	posc = _channels.find(",");

	while (posc >= 0 && !_channels.empty()) {
		range = _channels.substr(0, posc);
		posr = range.find("-");
		if (posr >= 0) {
			if (posr == 0) {
				fc = 1;
			} else {
				sscanf(_channels.substr(0, posr).c_str(), "%d", &fc);
			}
			sscanf(_channels.substr(posr+1, posc).c_str(), "%d", &lc);
		} else {
			sscanf(_channels.substr(0, posc).c_str(), "%d", &fc);
			lc = fc;
		}
		for (int i = fc; i <= lc; i++) {
			if (_channelnumbers->empty()) {
				_channelnumbers->push_back(i);
			} else {
				for (it = _channelnumbers->begin(); *it < i && it != _channelnumbers->end(); it++) {
				}
				if (it == _channelnumbers->end() && *it != i) {
					_channelnumbers->push_back(i);
				} else if (*it > i) {
					_channelnumbers->insert(it, i);
    }
  }
  }
		_channels.erase(0, posc + 1);

		posc = _channels.find(",");
	}
	if (!_channels.empty()) {
		range = _channels;
		posr= range.find("-");
		if (posr >= 0) {
			if (posr == 0) {
				fc = 1;
			} else {
				sscanf(_channels.substr(0, posr).c_str(), "%d", &fc);
			}
			sscanf(_channels.substr(posr + 1, posc).c_str(), "%d", &lc);
		} else {
			sscanf(_channels.c_str(), "%d", &fc);
			lc = fc;
		}
		for (int i = fc; i <= lc; i++) {
			if (_channelnumbers->empty()) {
				_channelnumbers->push_back(i);
			} else {
				for (it = _channelnumbers->begin(); *it < i && it != _channelnumbers->end(); it++) {
				}
				if (it == _channelnumbers->end() && *it != i) {
					_channelnumbers->push_back(i);
				} else if (*it > i) {
					_channelnumbers->insert(it, i);
				}
			}
		}
	}
}

Epg::Epg() :

	db((conf->p_homedir() + "epg.db").c_str()), exit_loop(false),
			x_start(1), epg_script_found(true),
			update_running(false) {

#ifdef use_nls
  // gettext
  setlocale(LC_ALL, ""); /* set from the environment variables */
  bindtextdomain("mms-epg", mms_prefix"/share/locale");
  bind_textdomain_codeset("mms-epg", nl_langinfo(CODESET));
#endif

	epg_conf = S_EPGConfig::get_instance();

	// Resolution-Independent-Stuff
	header_font     = "Vera/" + conv::itos(resolution_dependant_font_size(28, conf->p_v_res()));
	title_font      = "Vera/" + conv::itos(resolution_dependant_font_size(22, conf->p_v_res()));
	subtitle_font   = "Vera/" + conv::itos(resolution_dependant_font_size(18, conf->p_v_res()));
	normal_font     = "Vera/" + conv::itos(resolution_dependant_font_size(16, conf->p_v_res()));
	annotation_font = "Vera/" + conv::itos(resolution_dependant_font_size(12, conf->p_v_res()));
	grid_font	    = "Vera/" + conv::itos(resolution_dependant_font_size(14, conf->p_v_res()));;

	header_size 	= string_format::calculate_string_size("abcltuwHPMjJg", header_font);
	title_size  	= string_format::calculate_string_size("abcltuwHPMjJg", title_font);
	subtitle_size 	= string_format::calculate_string_size("abcltuwHPMjJg", subtitle_font);
	normal_size 	= string_format::calculate_string_size("abcltuwHPMjJg", normal_font);
	annotation_size = string_format::calculate_string_size("abcltuwHPMjJg", annotation_font);
    grid_size       = string_format::calculate_string_size("abcltuwHPMjJg", grid_font);

    channel_size    = string_format::calculate_string_size("abcltuwHPMjJgmi", grid_font);

	header_box_size = static_cast<int>(header_size.second * 0.75);


	grid_header_y = BORDER_TOP + header_size.second + SPACING + annotation_size.second;
	grid_spacing  = int(3 * conf->p_v_res() / 405); // for adjustment of vertical spacing on higher resolutions
	grid_y		  = grid_header_y + normal_size.second + SPACING;

	timeslot_interval    = (epg_conf->p_timeslot_interval() > 5 ? epg_conf->p_timeslot_interval() : 30);

	display_channel_name = epg_conf->p_display_channel_name();
	display_channel_logo = epg_conf->p_display_channel_logo();

	timeslots = epg_conf->p_timeslots();

	grid_header_x = int((conf->p_h_res() * 0.75) + SPACING/2);
	grid_x		  = grid_header_x;

	scale_factor = (timeslot_interval * SECONDS_PER_MINUTE * timeslots) / (double(conf->p_h_res() * 0.75) - (SPACING / 2) - BORDER_RIGHT);

	max_logo_size_x = 0;

	y_start = grid_y;

	rows = (conf->p_v_res() - (grid_y + BORDER_BOTTOM + normal_size.second + 2 * SPACING)) / (grid_size.second + grid_spacing);


	use_svdrp = false;
	lastrefresh = 0;
	pipe_is_nl = epg_conf->p_svdrp_pipe_is_nl();

	if (epg_conf->p_use_svdrp()) {
		svdrp_channelnumbers = new std::vector<int>;

		Timers = new std::vector<SvdrpTimerEvent>;
		Recordings = new std::vector<Recording>;

		svdrp = new Svdrp(epg_conf->p_svdrp_host(), epg_conf->p_svdrp_port());
		if (svdrp) {
			parse_channels(epg_conf->p_svdrp_channels(), svdrp_channelnumbers);
			timer_lead = epg_conf->p_svdrp_timer_lead();
			if (timer_lead <= 1) {
				timer_lead = 1;
			}
			timer_trailer = epg_conf->p_svdrp_timer_trailer();			
			if (timer_trailer <= 1) {
				timer_trailer = 1;
			}
			use_svdrp = true;

			if (epg_conf->p_svdrp_convert()) {
				svdrp->SetLocale(epg_conf->p_svdrp_locale());
			}
		}
	}

  epg_updater = new EpgUpdate();
}

Epg::~Epg() {
  delete epg_updater;

  db_mutex.enterMutex();
  db.close();
  db_mutex.leaveMutex();
}

bool Epg::valid_program() {
  bool found_match = false;

	time_t sel_start_time = (*program_iter)->start;

	vector<EpgEvent *>::iterator piter;
	for (piter = sel_channel->program.begin(), program_iter_end
			= sel_channel->program.end(); piter != program_iter_end; ++piter)
		if ((*piter)->stop > sel_start_time && (*piter)->stop > cur_time) {
      found_match = true;
      break;
    }

  program_iter = piter;

  return found_match;
}

void Epg::find_next_program() {
	vector<EpgEvent *>::iterator end = sel_channel->program.end();
	
  --end;

	vector<EpgEvent *>::iterator piter;
	for (piter = sel_channel->program.begin(), program_iter_end = sel_channel->program.end(); piter != program_iter_end; piter++) {
		if ((*piter)->start >= sel_time) {
			if ((*piter)->start == sel_time && piter != end) {

				vector<EpgEvent *>::iterator last_program_iter = piter;
	++piter;

	// fix for strange programs with no length
				if ((*last_program_iter)->start == (*last_program_iter)->stop
						|| (*piter)->start == (*piter)->stop && piter != end) {
	  ++piter;
	}
      }
      break;
    }
	}

	if (piter == program_iter_end) { // Epg-Info Ends too early
		if (sel_channel->program.size() > 0)
			piter--;
	}

	program_iter = piter;

	sel_time = (*program_iter)->start;

	int time = timeslots*timeslot_interval*SECONDS_PER_MINUTE;

  while (sel_time >= cur_time + time)
		cur_time += timeslot_interval * SECONDS_PER_MINUTE;
}

void Epg::find_valid_program() {
  if (!valid_program()) {
    find_next_program();
    return;
  }

	sel_time = (*program_iter)->start;

	int time = timeslots*timeslot_interval*SECONDS_PER_MINUTE;

  while (sel_time >= cur_time + time)
		cur_time += timeslot_interval*SECONDS_PER_MINUTE;
}

bool Epg::find_logos() {
	bool logos = true;

    std::vector<Channel>::iterator ch;
	
	foreach(Channel& channel, channels) {
		string logo_file = epg_conf->p_icon_path() + "/" + channel.display_name + ".xpm";
		if (file_exists(logo_file)) {
			channel.logo = logo_file;
			grid_header_x = int((conf->p_h_res() * 0.75) + SPACING/2);
			render->image_mut.enterMutex();
			render->device->wait_for_not_drawing();
			render->create_scaled_image_wrapper_upscaled(logo_file, int(conf->p_h_res() * 0.25 - SPACING/2), grid_size.second);
			
			PFObj *l = new PFObj(logo_file, 1, 1, int(conf->p_h_res() * 0.25 - SPACING/2), grid_size.second, false, 2);

			if (l) {
				if (l->w > max_logo_size_x)
					max_logo_size_x = l->w;

				delete l;
			}
			render->image_mut.leaveMutex();
			
		} else {
			logos = false;
			channel.logo = "";
			print_debug(dgettext("mms-epg", "logo not found:") + logo_file,
					"EPG");
		}
	}
	return logos;
}

bool Epg::check_tv_data() {
  // check if the TV.xml file is new enough
  time_t max = 0;

	if (!use_svdrp) {
		foreach(Channel& channel, channels) {
	
			string::size_type dot_pos = channel.id.find(".");
	
			if (dot_pos == string::npos) {
				dot_pos = channel.id.size();
				print_critical(dgettext("mms-epg",
						"Error in channel id, please report this problem"), "EPG");
			}
		}
	}

	foreach (Channel& channel, channels) {
		if (channel.program.size()> 0) {
			vector<EpgEvent *>::iterator cha = channel.program.end();
      --cha;
			if (cha != channel.program.begin())
	--cha;

			max = std::max((*cha)->stop, max);
		}
    }

  if (max < time(0)) {
		Print pdialog(dgettext("mms-epg", "Your TV Data is too old or inaccessible"),
		  Print::SCREEN);

    return false;
  } else
    return true;
}


// commands
void Epg::print_description() {
  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  S_BusyIndicator::get_instance()->idle();

	if ((*program_iter)->desc.size() != 0) {
    screen_updater->timer.deactivate("epg");
		(*program_iter)->PrintDescription(render, themes, conf->p_h_res(), conf->p_v_res());
    screen_updater->timer.activate("epg");
  } else
    DialogWaitPrint pdialog(dgettext("mms-epg", "No description available"), 1500);
}


bool Epg::toggle_timer() {
  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  S_BusyIndicator::get_instance()->idle();

  screen_updater->timer.deactivate("epg");

	if (use_svdrp) {
		if (!svdrp->Connect()) {
			screen_updater->timer.activate("epg");
			return false;
  }
		if (!svdrp->GetTimer(Timers)) {
			svdrp->Disconnect();
  screen_updater->timer.activate("epg");
			return false;
}
		if (!svdrp->AssociateTimersAndEpgEvents(&channels, Timers)) {
			svdrp->Disconnect();
			screen_updater->timer.activate("epg");
			return false;
		}
		svdrp->Disconnect();

		if ((*program_iter)->ToggleTimer(render, themes, conf->p_h_res(), conf->p_v_res())) {
			if ( (*program_iter)->HasTimer() ) {
				screen_updater->timer.activate("epg");
				return delete_timer((SvdrpEpgEvent *)*program_iter);
			} else {
				screen_updater->timer.activate("epg");
				return create_timer((SvdrpEpgEvent *)*program_iter);
			}		
		}
	}
}

bool Epg::switch_to_channel() {
    bool result;
    
    if (use_svdrp) {
        svdrp_mutex.enterMutex();
		if (!svdrp->Connect()) {
		    svdrp_mutex.leaveMutex();
			return false;
		}
	
		result = svdrp->SwitchtoChannel(sel_channel->number);
	
		svdrp->Disconnect();
	    svdrp_mutex.leaveMutex();
		if (result) {
			  run::exclusive_external_program(epg_conf->p_svdrp_tv_path() + ' ' + epg_conf->p_svdrp_tv_opts());
		}
		
		return result;
    }
}

bool Epg::create_timer(SvdrpEpgEvent *_epgevent) {
	bool result = false;

    svdrp_mutex.enterMutex();
	if (!svdrp->Connect()) {
		svdrp_mutex.leaveMutex();
		return false;
	}
	
	result = svdrp->NewTimer(TIMER_ACTIVE, sel_channel->number, "1", _epgevent->start - (timer_lead * SECONDS_PER_MINUTE), _epgevent->stop + (timer_trailer * SECONDS_PER_MINUTE), 
			50, 50, _epgevent->org_title, "<created by mms>");

	if (result) {
		for (std::vector<SvdrpTimerEvent>::iterator te=Timers->begin(); te != Timers->end(); te++) {
			if (te->epgevent != NULL) {
				((SvdrpEpgEvent *)(te->epgevent))->Cleartimer();
			}
		}
		
		Timers->clear();
		
		result = svdrp->GetTimer(Timers);
		
		if (result) {
			result = svdrp->AssociateTimersAndEpgEvents(&channels, Timers);
		}
	}

	svdrp->Disconnect();
	svdrp_mutex.leaveMutex();
	
    if (result) {
		Print pdialog(Print::SCREEN);
		pdialog.add_line(dgettext("mms-epg", "Timer created !"));
		pdialog.print();
	} else {
		Print pdialog(Print::SCREEN);
		pdialog.add_line(dgettext("mms-epg", "Timer not created !"));
		pdialog.add_line(svdrp->Message());
		pdialog.print();
	}

	return result;
}

bool Epg::delete_timer(SvdrpEpgEvent *_epgevent) {
    bool result = true;
    
	
	if (_epgevent->Timer()->IsRecording()) {
		svdrp_mutex.enterMutex();
		if (!svdrp->Connect()) {
			svdrp_mutex.leaveMutex();
			return false;
		}
	    ExtraMenu em(dgettext("mms-epg", "Timer is recording - really delete?"));
		em.add_item(ExtraMenuItem(dgettext("mms-epg", "Yes"), "", NULL));
		em.add_item(ExtraMenuItem(dgettext("mms-epg", "No"), "", NULL));
		int ret = em.mainloop();
		
		if (ret != 0) {
			svdrp->Disconnect();
			svdrp_mutex.leaveMutex();
			return true;
		}
		if (!svdrp->ModifyTimerOnOff(_epgevent->Timer()->Number(), false)) {
			result = false;
		}
		svdrp->Disconnect();
		svdrp_mutex.leaveMutex();
	}

	svdrp_mutex.enterMutex();
	if (!svdrp->Connect()) {
		svdrp_mutex.leaveMutex();
		return false;
	}

	if (result) {
		result = svdrp->DeleteTimer(_epgevent->Timer()->Number());
	
		if (result) {
			for (std::vector<SvdrpTimerEvent>::iterator te=Timers->begin(); te != Timers->end(); te++) {
				if (te->epgevent != NULL) {
					((SvdrpEpgEvent *)(te->epgevent))->Cleartimer();
				}
			}
			Timers->clear();
			
			result = svdrp->GetTimer(Timers);
			
			if (result) {
				result = svdrp->AssociateTimersAndEpgEvents(&channels, Timers);
			}
		}
	
		svdrp->Disconnect();
		svdrp_mutex.leaveMutex();
	}
	svdrp_mutex.leaveMutex();
	
	if (result) {
		Print pdialog(Print::SCREEN);
		pdialog.add_line(dgettext("mms-epg", "Timer deleted !"));
		pdialog.print();
	} else {
		Print pdialog(Print::SCREEN);
		pdialog.add_line(dgettext("mms-epg", "Timer not deleted !"));
		pdialog.add_line("");
		pdialog.add_line(svdrp->Message());
		pdialog.print();	
	}
	return result;
}

void Epg::exit() {
    input_master->restore_map();
  exit_loop = true;
}

bool Epg::move_element_up() {
  if (sel_channel != channels.begin()) {
    --sel_channel;

    if (sel_channel_num > 0) {
      --sel_channel_num;
    } else {
      --cur_channel;
    }

    return true;
  } else
    return false;
}

bool Epg::move_element_down() {
  if (sel_channel != (--channels.end())) {
    ++sel_channel;

    if (sel_channel_num < (rows - 1)) {
      ++sel_channel_num;
    } else {
      ++cur_channel;
    }

    return true;
  } else
    return false;
}

string Epg::mainloop() {
  Input input;

	input_master->save_map();
	input_master->set_map("epg");

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  // disable screensaver
  screen_updater->timer.deactivate("audio_fullscreen");

	screen_updater->timer.add(TimeElement("epg", boost::bind(
			&Epg::check_epg_print, this), boost::bind(&Epg::epg_print, this)));

  bool update_needed = true;

  BusyIndicator *busy_indicator = S_BusyIndicator::get_instance();
  Shutdown *sd = S_Shutdown::get_instance();

	refresh_mutex.enterMutex();
	program_iter = channels.begin()->program.begin();
	refresh_mutex.leaveMutex();

	while (!exit_loop) {
		if (update_needed) {
			refresh_mutex.enterMutex();
      print();
			refresh_mutex.leaveMutex();
		}

    input = input_master->get_input_busy_wrapped(busy_indicator);

    if (sd->is_enabled()) {
      	sd->cancel();
      	continue;
      }

    update_needed = true;

    if (input.key == "touch_input") {
      S_Touch::get_instance()->run_callback();
      continue;
    }

		if (input.command == "back" && input.mode == "general") {
	exit();
		} else if (input.command == "next") {
			refresh_mutex.enterMutex();
      if (move_element_down())
	find_valid_program();
			refresh_mutex.leaveMutex();
    } else if (input.command == "prev") {
			refresh_mutex.enterMutex();
      if (move_element_up())
	find_valid_program();
			refresh_mutex.leaveMutex();
    } else if (input.command == "page_up") {
			refresh_mutex.enterMutex();

      bool changed_channel = false;

      for (int i = 0; i < conf->p_jump(); ++i) {
	if (!move_element_up())
	  break;
	else
	  changed_channel = true;
      }

      if (changed_channel)
	find_valid_program();

			refresh_mutex.leaveMutex();

    } else if (input.command == "page_down") {
			refresh_mutex.enterMutex();

      bool changed_channel = false;

      for (int i = 0; i < conf->p_jump(); ++i) {
	if (!move_element_down())
	  break;
	else
	  changed_channel = true;
      }

      if (changed_channel)
	find_valid_program();

			refresh_mutex.leaveMutex();

    } else if (input.command == "action") {
			refresh_mutex.enterMutex();

      print_description();

			refresh_mutex.leaveMutex();

    } else if (input.command == "right") {
			refresh_mutex.enterMutex();

      find_next_program();

			refresh_mutex.leaveMutex();

    } else if (input.command == "left") {
			refresh_mutex.enterMutex();

			vector<EpgEvent *>::iterator old_program_iter = program_iter;
			vector<EpgEvent *>::iterator piter;

			for (piter = sel_channel->program.begin(); piter != sel_channel->program.end(); ++piter) {
				if ((*piter)->start > sel_time) {
					if (piter != sel_channel->program.begin())
	    --piter;
					if (piter != sel_channel->program.begin())
	    --piter;
					if ((*old_program_iter)->stop == (*piter)->stop	&& (*old_program_iter)->start == (*piter)->start)
						if (piter != sel_channel->program.begin())
	      --piter;

	  break;
	}
      }

      program_iter = piter;

			if (program_iter == sel_channel->program.end() && program_iter != sel_channel->program.begin()) {
	--program_iter;
			}
			if (program_iter == sel_channel->program.end() && program_iter != sel_channel->program.begin()) {
	--program_iter;
      }

			sel_time = (*program_iter)->start;

      while (sel_time < cur_time)
				cur_time -= timeslot_interval*SECONDS_PER_MINUTE;

			refresh_mutex.leaveMutex();

		} else if (input.command == "second_action") {
			ExtraMenu em;

			em.add_item(ExtraMenuItem(dgettext("mms-epg", "Print description"),
					input_master->find_shortcut("action"), boost::bind(
							&Epg::print_description, this)));

			em.add_item(ExtraMenuItem(dgettext("mms-epg", "Reload Data"),
					input_master->find_shortcut("reload"), boost::bind(
							&Epg::read_db, this)));
			if (use_svdrp) {
				em.add_item(ExtraMenuItem(dgettext("mms-epg", "Toggle Timer"),
						input_master->find_shortcut("toggle_timer"), boost::bind(
								&Epg::toggle_timer, this)));

				em.add_item(ExtraMenuItem(dgettext("mms-epg", "Switch to Channel"),
						input_master->find_shortcut("switch_to_channel"), boost::bind(
								&Epg::switch_to_channel, this)));
				
			}
			em.add_item(ExtraMenuItem(
					dgettext("mms-epg", "Return to startmenu"),
					input_master->find_shortcut("back"), boost::bind(
							&Epg::exit, this)));

	foreach (ExtraMenuItem& item, global->menu_items)
	  em.add_persistent_item(item);

	em.mainloop();
		} else if (input.command == "toggle_timer") {
			refresh_mutex.enterMutex();

			toggle_timer();
			refresh_mutex.leaveMutex();

		} else if (input.command == "reload") {
			read_db();
		} else if (input.command == "switch_to_channel") {
			switch_to_channel();
		} else
      update_needed = !global->check_commands(input);
  }

    input_master->restore_map();
    
  screen_updater->timer.del("epg");

  exit_loop = false;

  screen_updater->timer.activate("audio_fullscreen");

	refresh_mutex.leaveMutex();

  return "";
}

void Epg::print_timelines(double diff) {
	if (diff > 0 && diff < timeslots)
		render->current.add(new RObj(int(grid_x + diff * timeslot_interval * SECONDS_PER_MINUTE / scale_factor), grid_y - SPACING, 2, rows * (grid_size.second + grid_spacing) + SPACING * 2,
				themes->epg_time_bar1, themes->epg_time_bar2,
				themes->epg_time_bar3, themes->epg_time_bar4, 1));

	render->current.add(new RObj(grid_x, grid_y - SPACING, 2, rows * (grid_size.second + grid_spacing) + SPACING * 2,
			themes->epg_bar1, themes->epg_bar2,
			themes->epg_bar3, themes->epg_bar4, 1));

	for (int x = 0; x < timeslots; ++x) {
		time_t tmp_time = cur_time + x*timeslot_interval*SECONDS_PER_MINUTE;
		char time_str[30];
		strftime(time_str, 30, "%H:%M", localtime(&tmp_time));
		
		std::pair<int, int> time_size = string_format::calculate_string_size(time_str, grid_font);
		render->current.add(new TObj(time_str, grid_font,
				static_cast<int>(grid_x - (time_size.first / 2) + (x * timeslot_interval * SECONDS_PER_MINUTE)/scale_factor), grid_header_y ,
				themes->epg_font1, themes->epg_font2,
				themes->epg_font3, 2));
		render->current.add(new RObj(static_cast<int>(grid_x + ((x+1) * timeslot_interval * SECONDS_PER_MINUTE) / scale_factor), grid_y - SPACING, 2, rows * (grid_size.second + grid_spacing) + SPACING * 2,
				themes->epg_bar1, themes->epg_bar2, themes->epg_bar3,
				themes->epg_bar4, 1));
	}
}

void Epg::print_channel(int y, std::vector<Channel>::iterator cur_chan) {
	if (display_channel_logo) {
		if (!cur_chan->logo.empty()) {
			render->create_scaled_image_wrapper_upscaled(cur_chan->logo, max_logo_size_x, grid_size.second);
		
			render->current.add(new PFObj(cur_chan->logo, logo_x, grid_y + y * (grid_size.second + grid_spacing),  max_logo_size_x, grid_size.second, false, 2));
		}
	}
	if (display_channel_name) {
		string channel_name = cur_chan->display_name;
		string_format::format_to_size(channel_name, grid_font, grid_x - channel_x - SPACING/2, false);
		render->current.add(new TObj(channel_name, grid_font, channel_x, grid_y + y * (grid_size.second + grid_spacing) ,
				themes->epg_font1, themes->epg_font2,
				themes->epg_font3, 2));
	}
}

void Epg::print() {
	int y = BORDER_TOP;

	if (!all_logos) {
		display_channel_name = true;
	}
	
	if (display_channel_logo) {
		logo_x = BORDER_LEFT;
		channel_x = logo_x + max_logo_size_x + SPACING;
	} else {
		logo_x = 0;
		channel_x = BORDER_LEFT;
		display_channel_name = true;
	}

	if (display_channel_name == false) { // This only happens when all Channel-Logos are present and display_channel_name is set to false
		grid_header_x = BORDER_LEFT + max_logo_size_x + SPACING;
		grid_x = grid_header_x;
	} else {
		grid_header_x = int((conf->p_h_res() * 0.25) + SPACING/2);
		grid_x		  = grid_header_x;
	}
	
	scale_factor = (timeslot_interval * SECONDS_PER_MINUTE * timeslots) / (double(conf->p_h_res() - grid_x - BORDER_RIGHT));
	
  render->prepare_new_image();

  render->current.add(new PObj(themes->epg_background, 0, 0, 0, SCALE_FULL));

	PObj *back = new PFObj(themes->startmenu_epg, BORDER_LEFT, y, header_box_size, header_box_size, 1, true);

  render->current.add(back);

	S_Touch::get_instance()->register_area(TouchArea(
			rect(back->x, back->y, back->w, back->h), back->layer, boost::bind(
					&Epg::exit, this)));

  if (themes->show_header) {
		TObj *t = new TObj(dgettext("mms-epg", "EPG"), header_font, BORDER_LEFT + header_box_size + SPACING, y + (header_box_size - header_size.second) / 2,
		       themes->epg_header_font1, themes->epg_header_font2,
		       themes->epg_header_font3, 1);

    render->current.add(t);

		S_Touch::get_instance()->register_area(TouchArea(rect(t->x, t->y, t->w, t->h),
				t->layer, global->toggle_playlist));
  }

  char time_str[30];
  time_t tmp_time = time(0);
	strftime(time_str, 30, "%x - %R", localtime(&tmp_time));

	int x_size = string_format::calculate_string_width(time_str, normal_font);

  int time_pos_from_right = 15;

  if (conf->p_busy_indicator())
    time_pos_from_right = 32;

	render->current.add(new TObj(time_str, normal_font, conf->p_h_res()-(x_size + time_pos_from_right), y +(header_box_size - normal_size.second) / 2,
			       themes->epg_font1, themes->epg_font2,
			       themes->epg_font3, 2));

	y += header_size.second + SPACING;
	
	strftime(time_str, 30, "%x - %R", localtime(&cur_time));

	x_size = string_format::calculate_string_width(time_str, annotation_font);

	render->current.add(new TObj(time_str, annotation_font, (conf->p_h_res()-x_size)/2, y,
			themes->epg_font1, themes->epg_font2,
			themes->epg_font3, 2));


  time_t temp_time = time(0);

  double diff = 0;
	diff += (temp_time-cur_time)/(timeslot_interval*SECONDS_PER_MINUTE);
	diff += ((temp_time-cur_time)/SECONDS_PER_MINUTE-diff*timeslot_interval)/timeslot_interval;

  print_timelines(diff);

	if ((grid_x + timeslots * timeslot_interval * SECONDS_PER_MINUTE / scale_factor + BORDER_RIGHT) < conf->p_h_res()) {
		tmp_time = cur_time + timeslots*timeslot_interval*SECONDS_PER_MINUTE;
    strftime(time_str, 30, "%H:%M", localtime(&tmp_time));
		render->current.add(new TObj(time_str, grid_font, 
				int(grid_x - timeslots*timeslot_interval*SECONDS_PER_MINUTE/scale_factor),
				grid_header_y,
				 themes->epg_font1, themes->epg_font2, themes->epg_font3, 2));
  }

  std::ostringstream footer;
  int x, w;
  bool found_match = false;

	vector<Channel>::iterator cur_chan = cur_channel;

  for (int y = 0; y < std::min(rows, int(channels.size())); ++y) {
    print_channel(y, cur_chan);

		foreach (EpgEvent* program, cur_chan->program) {
			if (program->stop> cur_time &&
					program->start < (cur_time + timeslots*timeslot_interval*SECONDS_PER_MINUTE)) {

	// x = 80 + constant
				if (program->start < cur_time) {
					x = grid_x + 2;
					w = int((program->stop - cur_time)/scale_factor);
	} else {
					x = int(grid_x + 2 + (program->start - cur_time)/scale_factor);
					w = int((program->stop - program->start)/scale_factor);
	}

	if (w < 14) // very short shows
	  w = 14;

				if (x + w > grid_x + (timeslots*timeslot_interval*SECONDS_PER_MINUTE)/scale_factor) {
	  // programs ending at the end of screen should just be cut off, others
	  // should cross the line a little to indicate that they end later.
					w -= (x + w) - int(grid_x+(timeslots*timeslot_interval*SECONDS_PER_MINUTE)/scale_factor);
					if (program->stop> cur_time+timeslots*timeslot_interval*SECONDS_PER_MINUTE)
	    w += 15;
	}

				if (cur_chan == sel_channel && program->start <= sel_time &&
						program->stop> sel_time) {
	  if (!found_match) {

	    found_match = true;

						RObj *r = program->Print(render, themes, x, grid_y + y * (grid_size.second + grid_spacing), w, grid_size.second, true, grid_font);

	    footer.str("");
						strftime(time_str, 30, "%H:%M", localtime(&program->start));
	    footer << time_str << " - ";
						strftime(time_str, 30, "%H:%M", localtime(&program->stop));
	    footer << time_str << dgettext("mms-epg", " on ") << cur_chan->display_name;

						string f = footer.str() + "   " + program->title;

	    int f_size = f.size();
						string_format::format_to_size(f, grid_font, conf->p_h_res() - BORDER_LEFT - BORDER_RIGHT, false);

	    if (f.size() != f_size)
	      f += "...";

						render->current.add(new TObj(f, grid_font, BORDER_LEFT, grid_y + rows * (grid_size.second + grid_spacing) + (2 * SPACING), themes->epg_font1,
					 themes->epg_font2, themes->epg_font3, 4));

	    S_Touch::get_instance()->register_area(TouchArea(rect(r->x, r->y, r->w, r->h), r->layer,
										boost::bind(&Epg::print_description, this, program)));
	  }
	} else {

					RObj *r = program->Print(render, themes, x, grid_y + y * (grid_size.second + grid_spacing), w, grid_size.second, false, grid_font);

	  S_Touch::get_instance()->register_area(TouchArea(rect(r->x, r->y, r->w, r->h), r->layer,
									boost::bind(&Epg::print_description, this, program)));

	}
      }
    }
    ++cur_chan;
  }

  render->draw_and_release("epg");
}

void Epg::print_description(EpgEvent *p) {
	SvdrpEpgEvent *se;
	XmlEpgEvent *xe;
	
	se = (SvdrpEpgEvent *)p;
	xe = (XmlEpgEvent *)p;

	if (use_svdrp) {
		svdrp_mutex.enterMutex();
		if (svdrp->Connect()) {
			if (svdrp->GetTimer(Timers)) {
				svdrp->AssociateTimersAndEpgEvents(&channels, Timers);
			}
			svdrp->Disconnect();
		}
		svdrp_mutex.leaveMutex();
		
		se->PrintDescription(render, themes, conf->p_h_res(), conf->p_v_res());
	} else {
		xe->PrintDescription(render, themes, conf->p_h_res(), conf->p_v_res());	
	}
}

void Epg::epg_print() {
	if (conf->p_idle_time() != 0) {
		// update the idle timer
		int idle = time(0)-conf->p_last_key();
		if (idle > SECONDS_PER_MINUTE*conf->p_idle_time() && ((*program_iter)->stop > time(0) && (*program_iter)->start < (time(0) + timeslots*timeslot_interval*SECONDS_PER_MINUTE)))
			update_time();
	}

	
	if (use_svdrp) {
		svdrp_mutex.enterMutex();
		if (svdrp->Connect()) {
			if (svdrp->GetTimer(Timers)) {
				svdrp->AssociateTimersAndEpgEvents(&channels, Timers);
			}
			svdrp->Disconnect();
		}
		svdrp_mutex.leaveMutex();
	}

	print();
}

int Epg::check_epg_print()
{
	return SECONDS_PER_MINUTE*1000; // ms
}

void Epg::update_svdrp_data()
{
	 if (svdrp && !update_running) {
		 update_running = true;
		 DialogWaitPrint pdialog(dgettext("mms-epg", "Rereading TV Data"), 1000);
		 epg_updater->start();
	 }
}

void Epg::update_epg_datafile()
{
	 if (!epg_conf->p_epg_update_script().empty()
	       && file_exists(epg_conf->p_epg_update_script())
	       && !update_running) {
		 update_running = true;
	     epg_updater->start();
	 } else {
	     epg_script_found = false;
	     print_warning(dgettext("mms-epg", "Could not find epg update program"), "EPG");
	 }
}

void Epg::TimeStamp(std::string message) {
	char timestr[256];

	time_t now = time(NULL);
	struct tm *nowinfo = localtime(&now);

	strftime(timestr, 255, "%d.%m.%Y %H:%M:%S", nowinfo);

    printf("%s - Timestamp (%s)\n", timestr, message.c_str());		
}

bool Epg::read_xml_file(vector<Channel> *chan)
{
    std::vector<Channel>::iterator ch;
	std::vector<EpgEvent *> new_program, old_program;
	std::vector<EpgEvent *>::iterator e;

	TiXmlElement* root_element;
	TiXmlElement* element;

	refresh_mutex.enterMutex();

	if (!file_exists(epg_conf->p_epg_data())) {
		refresh_mutex.leaveMutex();
		return false;
	}

	TiXmlDocument doc(epg_conf->p_epg_data().c_str());

	doc.LoadFile();
	root_element = doc.RootElement();

	if (root_element == 0) {
		refresh_mutex.leaveMutex();
		return false;
    }

	if (channels.empty()) {

		channels.clear();

        int channel_number = 1;
	
		for (element = root_element->FirstChildElement("channel"); element != 0;
			element = element->NextSiblingElement("channel")) {
			Channel temp = XMLChannel(channel_number++, element);
			chan->push_back(temp);
		}

		for (ch = channels.begin(); ch != channels.end(); ch++) {
			if (ch->program.empty()) {
				time_t now = time(NULL);
				XmlEpgEvent *ev = new XmlEpgEvent(now, 7 * 24*60*SECONDS_PER_MINUTE, dgettext("mms-epg", "No Epg available"));
				ch->program.push_back(ev);
			}
		}

		cur_channel = channels.begin();
		sel_channel = cur_channel;
		sel_channel_num = 0;
	
		loaded_correctly = true;
		reload_needed = false;
		
		all_logos = find_logos();
		
		refresh_mutex.leaveMutex();
	} else {
		for (ch = channels.begin(); ch != channels.end(); ch++) {
			ch->program.clear();
			time_t now = time(NULL);
			XmlEpgEvent *ev = new XmlEpgEvent(now, 7 * 24*60*SECONDS_PER_MINUTE, dgettext("mms-epg", "No Epg available"));
			ch->program.push_back(ev);
		}

		cur_channel = channels.begin();
		sel_channel = cur_channel;
		sel_channel_num = 0;
	
		loaded_correctly = true;
		reload_needed = false;

		refresh_mutex.leaveMutex();		
	}
	
	string last_id = "";
	time_t min, max;

	for (element = root_element->FirstChildElement("programme"); element != 0; element = element->NextSiblingElement("programme")) {
		std::string channel_id = element->Attribute("channel");
		for (ch = channels.begin(); (ch != channels.end() && channel_id != ch->id); ch++) {
		}
		
		if (ch != channels.end()) { // Found matching Channel !!!

			
			XmlEpgEvent *p = new XmlEpgEvent(element);

			refresh_mutex.enterMutex();
			new_program = ch->program;
			old_program = ch->program;
			
			// Sort into vector
			e = new_program.begin();
			
			if ((*e)->title == dgettext("mms-epg", "No Epg available")) {
				new_program.clear();
			}

			for (e = new_program.begin(); e != new_program.end() && p->start > (*e)->start; e++) {
			}
			if (e == new_program.end()) {
				if (new_program.size() > 0) {
					e--;
					if ((*e)->stop == 0) {
						(*e)->stop = p->start;
					}
				}
				new_program.push_back(p);
			} else if (e == new_program.begin()) {
				new_program.insert(e, p);
				if (e != new_program.begin()) {
					e--;
					if ((*e)->stop == 0) {
						(*e)->stop = p->start;
					}
				}
			}
			if (min == 0)
				min = p->start;
			
			if (min > p->start)
				min = p->start;
			
			if (max == 0)
				max = (p->stop == 0 ? p->start : p->stop);

			if (max < p->stop)
				max = p->stop;
			
			if (ch == chan->begin()) {
				struct tm *tm;
			
				cur_time = time(0);
				sel_time = cur_time;
			
				for (program_iter = sel_channel->program.begin(), program_iter_end
							= sel_channel->program.end(); program_iter != program_iter_end; ++program_iter)
					if ((*program_iter)->stop >= sel_time)
	break;

				if (program_iter != program_iter_end) {
					cur_time = sel_time = (*program_iter)->start;
				}
    }
		}
		ch->program = new_program;


		old_program.clear();
		refresh_mutex.leaveMutex();
	}

	db_mutex.enterMutex();

	check_db();
	
	db.execute("BEGIN TRANSACTION");
	
	for (ch = channels.begin(); ch != channels.end(); ch++) {
		char *tmp = sqlite3_mprintf("delete from Channel where tvid = '%q'",  ch->id.c_str());
		db.execute(tmp);
		sqlite3_free(tmp);
		
		tmp = sqlite3_mprintf("delete from Program where cid = '%q'", ch->id.c_str());
		db.execute(tmp);
		sqlite3_free(tmp);
		
		tmp = sqlite3_mprintf("INSERT INTO Channel (id, number, tvid, name) VALUES (NULL, %d, '%q', '%q')",
			  ch->number, ch->id.c_str(), ch->display_name.c_str());
		db.execute(tmp);
		sqlite3_free(tmp);

		XmlEpgEvent *x;
		
		for (e = ch->program.begin(); e != ch->program.end(); e++) {
			x = (XmlEpgEvent *)(*e);	
			// FIXME: determine if movie, episode or ordinary program

			char *insert = sqlite3_mprintf("INSERT INTO Program (id, cid, title, ltitle, description, ldescription, start_time, end_time, duration) VALUES (NULL, '%q', '%q', '%q', '%q', '%q', %ld, %ld, %ld)",
					  ch->id.c_str(), x->title.c_str(),
					  string_format::lowercase(x->title).c_str(),
					  x->desc.c_str(), string_format::lowercase(x->desc).c_str(),
					  x->start, x->stop, x->duration);
	
			db.execute(insert);
			sqlite3_free(insert);
  }
}

	db.execute("END TRANSACTION");

	db_mutex.leaveMutex();

	return true;
  }

bool Epg::GetSvdrpData() {
    std::vector<Channel>::iterator ch;
	std::vector<EpgEvent *> new_program, old_program;
	
	refresh_mutex.enterMutex();

	svdrp_mutex.enterMutex();
	if (!svdrp->Connect()) {
		refresh_mutex.leaveMutex();
		svdrp_mutex.leaveMutex();
		return false;
}

	if (channels.empty()) {
		
		channels.clear();
		Timers->clear();

		if (!svdrp->ListChannels(svdrp_channelnumbers, &channels)) {
			svdrp->Disconnect();
			refresh_mutex.leaveMutex();
			svdrp_mutex.leaveMutex();
			return false;
}

		for (ch = channels.begin(); ch != channels.end(); ch++) {
			if (ch->program.empty()) {
				time_t now  = time(NULL);
				SvdrpEpgEvent *ev = new SvdrpEpgEvent(1, ch->id, now, 7 * 24*60*SECONDS_PER_MINUTE, 0, 0);
				ev->title = dgettext("mms-epg", "No Epg available");
				
				ch->program.push_back(ev);
			}
		}

		cur_channel = channels.begin();
		sel_channel = cur_channel;
		sel_channel_num = 0;

		loaded_correctly = true;
		reload_needed = false;
		
		// all_logos = false;
		
		all_logos = find_logos();
		
		refresh_mutex.leaveMutex();
  } else {
		refresh_mutex.leaveMutex();		
  }

	for (ch = channels.begin(); ch != channels.end(); ch++) {
		new_program = svdrp->GetEpgEvents(ch->number);

		if (!new_program.empty()) {
			refresh_mutex.enterMutex();
			old_program = ch->program;
			ch->program = new_program;
	
			old_program.clear();
			
			if (ch == channels.begin()) {
				struct tm *tm;
		
				cur_time = time(0);
				sel_time = cur_time;
		
				for (program_iter = sel_channel->program.begin(), program_iter_end
						= sel_channel->program.end(); program_iter != program_iter_end; ++program_iter)
					if ((*program_iter)->stop >= sel_time)
						break;

				if (program_iter != program_iter_end) {
					cur_time = sel_time = (*program_iter)->start;
				}
			}
			refresh_mutex.leaveMutex();
		}
	}

	if (!svdrp->GetTimer(Timers)) {
		svdrp->Disconnect();
		svdrp_mutex.leaveMutex();
		return false;
	}

	if (!svdrp->AssociateTimersAndEpgEvents(&channels, Timers)) {
		svdrp->Disconnect();
		svdrp_mutex.leaveMutex();
		return false;
	}
	svdrp->Disconnect();

	svdrp_mutex.leaveMutex();

	// Write Stuff to Database
	
	db_mutex.enterMutex();

	check_db();
	
	db.execute("BEGIN TRANSACTION");

	for (ch = channels.begin(); ch != channels.end(); ch++) {
		SvdrpChannel *sc = (SvdrpChannel *)&(*ch);
		
		char *tmp = sqlite3_mprintf("delete from Channel where tvid = '%q'",  sc->id.c_str());
		db.execute(tmp);
		sqlite3_free(tmp);

		tmp = sqlite3_mprintf("delete from Program where cid = '%q'", sc->id.c_str());

		db.execute(tmp);
		sqlite3_free(tmp);

		tmp = sqlite3_mprintf("INSERT INTO Channel (id, number, tvid, name) VALUES (NULL, %d, '%q', '%q')", 
			  sc->number, 
			  sc->id.c_str(), 
			  sc->display_name.c_str()
			  );

		db.execute(tmp);
		sqlite3_free(tmp);
		
		std::vector<EpgEvent *>::iterator ev;
		SvdrpEpgEvent *p;
		
		for (ev = ch->program.begin(); ev != ch->program.end(); ev++) {
			p = (SvdrpEpgEvent *)(*ev);		
			char *insert = sqlite3_mprintf("INSERT INTO Program (id, cid, title, ltitle, description, ldescription, start_time, end_time, duration, vps) VALUES (NULL, '%q', '%q', '%q', '%q', '%q', %ld, %ld, %ld, %ld)",
					  ch->id.c_str(), p->title.c_str(),
					  string_format::lowercase(p->title).c_str(),
					  p->desc.c_str(), string_format::lowercase(p->desc).c_str(),
					  p->start, p->stop, p->duration, p->Vps());
	
			db.execute(insert);
			sqlite3_free(insert);
		}
	}
	db.execute("END TRANSACTION");
	db_mutex.leaveMutex();
	
	return true;
}

void EpgUpdate::run()
{
  EPGConfig *epg_conf = S_EPGConfig::get_instance();

  Epg *epg = get_class<Epg>(dgettext("mms-epg", "Electronic Program Guide"));

  if (epg->use_svdrp) {
	  if (epg->svdrp) {
		  if (epg->GetSvdrpData()) {
			  Print pdialog(Print::SCREEN);
			  pdialog.add_line(dgettext("mms-epg", "EPG TV listings reloaded from Svdrp-Host"));
			  pdialog.print();
			  epg->reload_needed = false;
		  } else {
			  Print pdialog(Print::SCREEN);
			  pdialog.add_line(dgettext("mms-epg", "EPG TV listings couldn't be loaded from Svdrp-Host"));
			  pdialog.add_line("");
			  pdialog.print();	  
			  epg->reload_needed = true;
		  }
		  epg->update_running = false;
	  }
  } else {
  run::external_program(epg_conf->p_epg_update_script());

  Print pdialog(Print::SCREEN);
  pdialog.add_line(dgettext("mms-epg", "EPG TV listings file reloaded"));
  pdialog.add_line("");
  pdialog.add_line(dgettext("mms-epg", "The new data file will be used"));
  pdialog.add_line(dgettext("mms-epg", "the next time you enter EPG"));
  pdialog.print();

  epg->reload_needed = false;
  epg->update_running = false;
}
}

int Epg::check_epg_datafile() {
	time_t now = time(0);

	if (!epg_conf->p_epg_data().empty()) {
		if ((now - lastrefresh > (interval * 60)) && !update_running) {
    struct stat64 s;

    stat64(epg_conf->p_epg_data().c_str(), &s);

			time_t diff = s.st_mtime - lastrefresh;

			if (diff > 0) {
				lastrefresh = now;
				return 0;
			}
		}
		return (epg_conf->p_interval() >= 5) ? (epg_conf->p_interval()
					* 60 * 1000)
					: (5 * 60 * 1000); // ms
	} else
		return 24*60*60*1000; // just wait a long time
}

int Epg::check_svdrp_data() {
	time_t now = time(0);

	if (svdrp) {
		if ((now - lastrefresh > (interval * 60)) && !update_running) {
			lastrefresh = now;
      return 0; // update now
  }
		return (epg_conf->p_interval() >= 5) ? (epg_conf->p_interval()
				* 60 * 1000)
				: (5 * 60 * 1000); // ms
	} else
		return 24*60*60*1000; // Wait a day
}

void Epg::startup_updater() {
  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

	if (use_svdrp) {
		screen_updater->timer.add(TimeElement("epg_data_update",
				boost::bind(&Epg::check_svdrp_data, this),
				boost::bind(&Epg::update_svdrp_data, this)));
	} else {
		screen_updater->timer.add(TimeElement("epg_data_update", 
				boost::bind(&Epg::check_epg_datafile, this),
				boost::bind(&Epg::update_epg_datafile, this)));
		S_Search::get_instance()->register_module(SearchModule(dgettext("mms-epg", "EPG"),
				boost::bind(&Epg::gen_search_list, this, _1),
							 boost::bind(&Epg::reenter, this, _1)));
}
}

void Epg::reenter(int db_id) {
  if (!loaded_correctly) {
    read_db();
    loaded_correctly = true;
  }

	if (!check_tv_data()) {
		
    reload_needed = true;
    return;
  }

  db_mutex.enterMutex();

  SQLQuery *q = db.query("Program", ("SELECT cid, start_time, end_time FROM %t WHERE id='" + conv::itos(db_id) + "'").c_str());

  if (q && q->numberOfTuples() > 0) {

    SQLRow &row = (*q)[0];
    int start = conv::atoi(row["start_time"]);
    string tvid = row["cid"];

    int num = 0;
		for (std::vector<Channel>::iterator i = channels.begin(), iend =
				channels.end(); i != iend; ++i, ++num)
      if (i->id == tvid) {
	sel_channel = i;
	break;
      }

		for (program_iter = sel_channel->program.begin(), program_iter_end
				= sel_channel->program.end(); program_iter
				!= program_iter_end; ++program_iter)
			if ((*program_iter)->start == start)
	break;

		cur_time = sel_time = (*program_iter)->start;

    if (channels.size() > rows) {
      sel_channel_num = rows-1;

      int start_num = num - sel_channel_num, counter = 0;
			for (std::vector<Channel>::iterator i = channels.begin(), iend =
					channels.end(); i != iend; ++i, ++counter)
	if (counter == start_num) {
	  cur_channel = i;
	  break;
	}

    } else {
      sel_channel_num = num;
      cur_channel = channels.begin();
    }

    mainloop();
  }

  delete q;

  db_mutex.leaveMutex();
}

std::vector<std::pair<std::string, int> > Epg::gen_search_list(const std::string& search_word) {
  std::vector<std::pair<std::string, int> > result;

  db_mutex.enterMutex();

  // check if anything has been inserted, the db doesn't get created until one
  // has entered EPG

  bool tables_found = false;

	SQLQuery *dq = db.query("sqlite_master",
			"SELECT name FROM %t WHERE type='table'");
  if (dq) {
    for (int i = 0; i < dq->numberOfTuples(); ++i) {
      tables_found = true;
      break;
    }
  }

  delete dq;

  if (!tables_found)
    return result;

  SQLQuery *q = db.query("Program", ("SELECT DISTINCT * FROM %t WHERE fuzzycmp('" + search_word + "', ltitle, 1)").c_str());
  if (q) {
    for (int i = 0; i < q->numberOfTuples(); ++i) {
      SQLRow &row = (*q)[i];

      string out = row["title"];

      time_t start = conv::atoi(row["start_time"]);

      char time_str[30];
      strftime(time_str, 30, "%H:%M %d-%m-%Y", localtime(&start));

      SQLQuery *q2 = db.query("Channel", ("SELECT name FROM %t WHERE tvid='" + row["cid"] + "'").c_str());
      if (q2 && q2->numberOfTuples() > 0) {
	SQLRow &c_row = (*q2)[0];
	out += string(" ") + time_str + " on " + c_row["name"];
      }
      delete q2;

      result.push_back(std::make_pair(out, conv::atoi(row["id"])));
    }
  }
  delete q;

  db_mutex.leaveMutex();

  return result;
}
